

#include <stdio.h>
#include <mpi.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

int main(int argc, char *argv[])
{
  int npes, myrank;
  int p_rank, i, n, N;
  clock_t start, diff;
  char *msg;
  double timer, fin_timer;

  MPI_Init(&argc, &argv);

  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  if(myrank==0) {
    if(npes&1) {
      fprintf(stderr, "Number of processes must be an even number. \n");
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }
    
    if(argc < 3) {
      fprintf(stderr, "Give N, message as arguments.\n");
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }
  }

  N = atoi(argv[1]);
  msg = argv[2];
  n = strlen(msg);

  /* find partner rank */
  p_rank = (myrank&1) ? myrank-1 :  myrank+1;
    
  for(i = 0; i < N; i++) {
    start = clock();

    if(myrank&1) {
      MPI_Send(&msg, n, MPI_CHAR, p_rank, 0, \
	       MPI_COMM_WORLD);
      MPI_Recv(&msg, n, MPI_CHAR, p_rank, 0, \
	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    } else {
      MPI_Recv(&msg, n, MPI_CHAR, p_rank, 0, \
	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Send(&msg, n, MPI_CHAR, p_rank, 0, \
	       MPI_COMM_WORLD);
    }

    diff = clock() - start;
    timer+=diff;
  }

  MPI_Reduce(&timer, &fin_timer, 1, MPI_DOUBLE, \
	     MPI_SUM, 0, MPI_COMM_WORLD);

  fin_timer/=(N*npes);

  if(myrank==0)
    printf("Average overhead: %f ms\n", fin_timer);

  MPI_Finalize();

  return 0;
}
