
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

int main(int argc, char *argv[])
{
  int myrank, npes, M, my_M, lower;
  double x, dx;
  double t, dt;
  double *u, *up, *um, *tmp;
  int i, l_rank, r_rank;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  if(myrank==0) {
    if(argc < 2) {
      fprintf(stderr, "Give M as an argument.\n");
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }
  }

  /* calculating m, local m */
  M = atoi(argv[1]);
  my_M = (M-2)/npes + (((M-2)%npes)>myrank);
  dx = 1.0/M;
  dt = dx;

  /* calculating lower limit */
  lower = 0;
  for(i = 0; i < myrank; i++)
    lower += ((M-2)/npes + (((M-2)%npes)>i));

  /* allocating three 1D arrays um, u, up of length M+2 */
  um = malloc((my_M+2)*sizeof(double));
  u = malloc((my_M+2)*sizeof(double));
  up = malloc((my_M+2)*sizeof(double));

  /* independent calculations: can be computed locally */
  for (i=0; i< my_M+2; i++) {
    x = (lower+i)*dx;
    um[i] = sin(2.0*M_PI*x);
  }

  for (i=1; i<=my_M; i++)
    u[i] = um[i] + 0.5*(um[i-1]-2*um[i]+um[i+1]); 

  /* overlapping values */
  double l, r;
  l = sin(2.0*M_PI*(lower-1)*dx);
  r = sin(2.0*M_PI*(lower+my_M+2)*dx);

  if(myrank != 0)
    u[0] = 0.5*(l+um[1]);
  else
    u[0] = 0;

  if(myrank != npes -1)
    u[my_M+1] = 0.5*(r+um[my_M]);
  else
    u[my_M+1] = 0;

  /* calculating left/right rank for mpi_sendrecv */
  l_rank = myrank ? myrank - 1 : MPI_PROC_NULL;
  r_rank = myrank < npes-1 ? myrank + 1 : MPI_PROC_NULL;

  t = dt;
  while (t < 1.0) {
    t += dt;

    for (i = 1; i <= my_M; i++)
      up[i] = um[i] + u[i-1] + u[i+1];

    /* exchange overlapped regions */
    MPI_Sendrecv(&up[1],      1, MPI_DOUBLE, l_rank, 1,
		 &up[my_M+1], 1, MPI_DOUBLE, r_rank, 1,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    MPI_Sendrecv(&up[my_M], 1, MPI_DOUBLE, r_rank, 2,
		 &up[0],      1, MPI_DOUBLE, l_rank, 2,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    /* up[0] = up[size-1] */
    if (npes > 1) {
      if (myrank == 0)
	MPI_Recv(&up[0], 1, MPI_DOUBLE, myrank-1, 3,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

      if (myrank == npes-1)
	MPI_Send(&up[my_M+1], 1, MPI_DOUBLE, 0, 3,
		 MPI_COMM_WORLD);
    } else {
      up[0] = up[my_M+1];
    }
    /* shuffle the three arrays */
    tmp = um;
    um = u;
    u = up;
    up = tmp;
  }

  /* print values
  if(npes > 1) {
    if(myrank==0)
      printf("%d %f\n", myrank, u[0]);
    for(i = 1; i <= my_M; i++)
      printf("%d %f\n", myrank, u[i]);
    if(myrank==npes-1)
      printf("%d %f", myrank, u[my_M+1]);
  } else {
    for(i = 0; i < my_M+2; i++) {
      printf("%f\n", u[i]);
    }
  } 
  */

  free(u), free(up), free(um);

  MPI_Finalize();

  return 0;
}
